"""Utility function for computing scaling factors for non-dimensionalization."""

import numpy as np
from scipy.constants import elementary_charge, mu_0

from desc.geometry import Curve


def compute_scaling_factors(thing):
    """Compute dimensional quantities for normalizations."""
    # local import to avoid circular import
    from desc.equilibrium import Equilibrium
    from desc.geometry import FourierRZToroidalSurface

    scales = {}

    def get_lowest_mode(basis, coeffs):
        """Return lowest order coefficient (excluding m=0 modes)."""
        # lowest order modes: [0, +1, -1, +2, -2, ...]
        m_modes = np.arange(1, thing.M + 1)
        m_modes = np.vstack((m_modes, -m_modes)).flatten(order="F")
        n_modes = np.arange(thing.N + 1)
        n_modes = np.vstack((n_modes, -n_modes)).flatten(order="F")
        for n in n_modes:
            for m in m_modes:
                try:
                    x = coeffs[basis.get_idx(M=m, N=n)]
                    if not np.isclose(x, 0):  # mode exists and coefficient is non-zero
                        return x
                except ValueError:
                    pass
        raise ValueError("No modes found, geometry is unphysical.")

    if isinstance(thing, Equilibrium):
        R00 = thing.Rb_lmn[thing.surface.R_basis.get_idx(M=0, N=0)]
        R10 = get_lowest_mode(thing.surface.R_basis, thing.Rb_lmn)
        Z10 = get_lowest_mode(thing.surface.Z_basis, thing.Zb_lmn)

        scales["R0"] = R00
        scales["a"] = np.sqrt(np.abs(R10 * Z10))
        scales["Psi"] = abs(thing.Psi)
        scales["A"] = np.pi * scales["a"] ** 2
        scales["V"] = 2 * np.pi * scales["R0"] * scales["A"]
        scales["B"] = scales["Psi"] / scales["A"] * 1.25
        B_pressure = scales["B"] ** 2 / (2 * mu_0)
        scales["I"] = scales["B"] * 2 * np.pi / mu_0
        scales["W"] = B_pressure * scales["V"]
        scales["J"] = scales["B"] / scales["a"] / mu_0
        scales["F"] = B_pressure / scales["a"]
        scales["f"] = scales["F"] * scales["V"]

        if thing.pressure is not None:
            p0 = float(thing.pressure(0)[0])
        else:
            scales["n"] = float(
                ((thing.atomic_number(0) + 1) / 2 * thing.electron_density(0))[0]
            )
            scales["T"] = np.mean(
                [thing.electron_temperature(0), thing.ion_temperature(0)]
            )
            p0 = elementary_charge * 2 * scales["n"] * scales["T"]
        if p0 < 1:  # vacuum
            scales["p"] = B_pressure
        else:
            scales["p"] = p0
        scales["W_p"] = scales["p"] * scales["V"] / 2

    elif isinstance(thing, FourierRZToroidalSurface):
        R00 = thing.R_lmn[thing.R_basis.get_idx(M=0, N=0)]
        R10 = get_lowest_mode(thing.R_basis, thing.R_lmn)
        Z10 = get_lowest_mode(thing.Z_basis, thing.Z_lmn)

        scales["R0"] = R00
        scales["a"] = np.sqrt(np.abs(R10 * Z10))
        scales["A"] = np.pi * scales["a"] ** 2
        scales["V"] = 2 * np.pi * scales["R0"] * scales["A"]

    elif isinstance(thing, Curve):
        scales["a"] = thing.compute("length")["length"] / (2 * np.pi)

    # replace 0 scales to avoid normalizing by zero
    for scale in scales.keys():
        if np.isclose(scales[scale], 0):
            scales[scale] = 1

    return scales



"""Base classes for optimizable objects."""

import inspect
import warnings
from abc import ABC

import numpy as np

from desc.backend import jnp


class Optimizable(ABC):
    """Base class for all objects in DESC that can be optimized.

    Sub-classes should decorate optimizable attributes with the
    ``optimizable_parameter`` decorator.
    """

    @property
    def optimizable_params(self):
        """list: string names of parameters that have been declared optimizable."""
        if not hasattr(self, "_optimizable_params"):
            p = []
            for methodname in dir(self):
                if methodname.startswith("__"):
                    continue
                # to avoid executing code and causing recursion
                method = inspect.getattr_static(self, methodname)
                if isinstance(method, property):
                    method = method.fget  # we want the property itself, not the value
                if hasattr(method, "optimizable"):
                    p.append(methodname)
            self._optimizable_params = self._sort_args(p)
            if not len(p):
                warnings.warn(
                    f"Object {self} was subclassed from Optimizable but no "
                    + "optimizable parameters were declared"
                )
        return self._optimizable_params

    @property
    def params_dict(self):
        """dict: dictionary of arrays of optimizable parameters."""
        return {
            key: jnp.atleast_1d(jnp.asarray(getattr(self, key))).copy()
            for key in self.optimizable_params
        }

    @params_dict.setter
    def params_dict(self, d):
        for key, val in d.items():
            if jnp.asarray(val).size:
                setattr(self, key, val)

    @property
    def dimensions(self):
        """dict: dictionary of integers of sizes of each optimizable parameter."""
        return {
            key: jnp.asarray(getattr(self, key)).size for key in self.optimizable_params
        }

    @property
    def x_idx(self):
        """dict: arrays of indices for each parameter in concatenated array."""
        dimensions = self.dimensions
        idx = {}
        dim_x = 0
        for arg in self.optimizable_params:
            idx[arg] = jnp.arange(dim_x, dim_x + dimensions[arg])
            dim_x += dimensions[arg]
        return idx

    @property
    def dim_x(self):
        """int: total number of optimizable parameters."""
        return sum(self.dimensions.values())

    def pack_params(self, p):
        """Convert a dictionary of parameters into a single array.

        Parameters
        ----------
        p : dict
            Dictionary of ndarray of optimizable parameters.

        Returns
        -------
        x : ndarray
            optimizable parameters concatenated into a single array, with indices
            given by ``x_idx``

        """
        return jnp.concatenate(
            [jnp.atleast_1d(jnp.asarray(p[key])) for key in self.optimizable_params]
        )

    def unpack_params(self, x):
        """Convert a single array of concatenated parameters into a dictionary.

        Parameters
        ----------
        x : ndarray
            optimizable parameters concatenated into a single array, with indices
            given by ``x_idx``

        Returns
        -------
        p : dict
            Dictionary of ndarray of optimizable parameters.

        """
        x_idx = self.x_idx
        params = {}
        for arg in self.optimizable_params:
            params[arg] = jnp.atleast_1d(jnp.asarray(x[x_idx[arg]]))
        return params

    def _sort_args(self, args):
        """Put arguments in a canonical order. Returns unique sorted elements.

        Actual order doesn't really matter as long as its consistent, though subclasses
        may override this method to enforce a specific ordering.

        """
        return sorted(set(list(args)))


class OptimizableCollection(Optimizable):
    """Base class for collections of multiple optimizable objects (coilsets, etc).

    Subclasses should be iterable, where each member is itself Optimizable.
    """

    @property
    def optimizable_params(self):
        """list: string names of parameters that have been declared optimizable."""
        return [s.optimizable_params for s in self]

    @property
    def params_dict(self):
        """list: list of dictionary of arrays of optimizable parameters."""
        return [s.params_dict for s in self]

    @params_dict.setter
    def params_dict(self, d):
        for s, p in zip(self, d):
            s.params_dict = p

    @property
    def dimensions(self):
        """list: list of dictionary of integers of sizes of each parameter."""
        return [s.dimensions for s in self]

    @property
    def x_idx(self):
        """list: list of dict of arrays of idx for each param in concatenated array."""
        x_idx = [s.x_idx for s in self]
        offset = jnp.concatenate(
            [jnp.array([0]), jnp.cumsum(jnp.array([s.dim_x for s in self]))[:-1]]
        )
        for d, idx in zip(offset, x_idx):
            # offset subsequent indices by length of priors
            for key in idx:
                idx[key] += d
        return x_idx

    @property
    def dim_x(self):
        """int: total number of optimizable parameters."""
        return sum(s.dim_x for s in self)

    def pack_params(self, params):
        """Convert a list of dictionary of parameters into a single array.

        Parameters
        ----------
        params : list of dict
            list of dictionary of ndarray of optimizable parameters.

        Returns
        -------
        x : ndarray
            optimizable parameters concatenated into a single array, with indices
            given by ``x_idx``

        """
        return jnp.concatenate([s.pack_params(p) for s, p in zip(self, params)])

    def unpack_params(self, x):
        """Convert a single array of concatenated parameters into a dictionary.

        Parameters
        ----------
        x : ndarray
            optimizable parameters concatenated into a single array, with indices
            given by ``x_idx``

        Returns
        -------
        p : list dict
            list of dictionary of ndarray of optimizable parameters.

        """
        split_idx = np.cumsum([s.dim_x for s in self])  # must be np not jnp
        xs = jnp.split(x, split_idx)
        params = [s.unpack_params(xi) for s, xi in zip(self, xs)]
        return params


def optimizable_parameter(f):
    """Decorator to declare an attribute or property as optimizable.

    The attribute should be a scalar or ndarray of floats.

    Examples
    --------
    .. code-block:: python

        class MyClass(Optimizable):

            def __init__(self, x, y):
                self.x = x
                self.y = optimizable_parameter(y)

            @optimizable_parameter
            @property
            def x(self):
                return self._x

            @x.setter
            def x(self, new):
                assert len(x) == 10
                self._x = x

    """
    if isinstance(f, property):
        f.fget.optimizable = True
    else:
        f.optimizable = True
    return f



"""Tests for profile classes."""

import numpy as np
import pytest
from scipy.constants import elementary_charge
from scipy.interpolate import interp1d

from desc.equilibrium import Equilibrium
from desc.examples import get
from desc.grid import LinearGrid
from desc.io import InputReader
from desc.objectives import (
    ForceBalance,
    ObjectiveFunction,
    get_fixed_boundary_constraints,
)
from desc.profiles import (
    FourierZernikeProfile,
    HermiteSplineProfile,
    MTanhProfile,
    PowerSeriesProfile,
    SplineProfile,
    TwoPowerProfile,
)

from .utils import area_difference, compute_coords


class TestProfiles:
    """Tests for Profile classes."""

    @pytest.mark.slow
    @pytest.mark.regression
    @pytest.mark.solve
    def test_same_result(self):
        """Test that different representations of the same profile give the same eq."""
        input_path = "./tests/inputs/SOLOVEV"
        ir = InputReader(input_path)

        eq1 = Equilibrium(**ir.inputs[-1], check_kwargs=False)
        eq2 = eq1.copy()
        eq2.pressure = eq1.pressure.to_spline()
        eq2.iota = eq1.iota.to_spline()

        eq1.solve()
        eq2.solve()

        Rr1, Zr1, Rv1, Zv1 = compute_coords(eq1, Nz=6)
        Rr2, Zr2, Rv2, Zv2 = compute_coords(eq2, Nz=6)
        rho_err, theta_err = area_difference(Rr1, Rr2, Zr1, Zr2, Rv1, Rv2, Zv1, Zv2)
        np.testing.assert_allclose(rho_err, 0, atol=1e-7)
        np.testing.assert_allclose(theta_err, 0, atol=2e-11)

    @pytest.mark.unit
    def test_close_values(self):
        """Test that different forms of the same profile give similar values."""
        pp = PowerSeriesProfile(
            modes=np.array([0, 2, 4]), params=np.array([1, -2, 1]), sym=False
        )
        sp = pp.to_spline()
        with pytest.warns(UserWarning):
            mp = pp.to_mtanh(order=4, ftol=1e-12, xtol=1e-12)
        zp = pp.to_fourierzernike()
        # don't test to rho=1 bc mtanh is very non-polynomial there,
        # don't test at axis bc splines and mtanh don't enforce zero slope exactly.
        x = np.linspace(0.1, 0.8, 10)

        np.testing.assert_allclose(pp(x), sp(x), rtol=1e-5, atol=1e-3)
        np.testing.assert_allclose(pp(x, dr=2), mp(x, dr=2), rtol=1e-2, atol=1e-1)
        np.testing.assert_allclose(pp(x), zp(x), rtol=1e-5, atol=1e-3)
        np.testing.assert_allclose(pp(x, dr=2), zp(x, dr=2), rtol=1e-2, atol=1e-1)

        pp1 = sp.to_powerseries(order=4)
        np.testing.assert_allclose(pp.params, pp1.params, rtol=1e-5, atol=1e-2)
        pp2 = mp.to_powerseries(order=4)
        np.testing.assert_allclose(pp.params, pp2.params, rtol=1e-5, atol=1e-2)

        with pytest.warns(UserWarning):
            mp1 = sp.to_mtanh()
        np.testing.assert_allclose(mp1(x, dr=1), sp(x, dr=1), rtol=1e-5, atol=1e-3)

        pp3 = pp.to_powerseries()
        sp3 = sp.to_spline()
        np.testing.assert_allclose(sp3(x), pp3(x), rtol=1e-5, atol=1e-3)
        sp4 = mp.to_spline()
        np.testing.assert_allclose(sp3(x), sp4(x), rtol=1e-5, atol=1e-2)

        np.testing.assert_allclose(pp(x, dt=1), 0)
        np.testing.assert_allclose(pp(x, dz=1), 0)
        np.testing.assert_allclose(sp(x, dt=1), 0)
        np.testing.assert_allclose(sp(x, dz=1), 0)
        np.testing.assert_allclose(mp(x, dt=1), 0)
        np.testing.assert_allclose(mp(x, dz=1), 0)

        pp = PowerSeriesProfile([0.5, 0, -1, 0, 0.5])
        tp = TwoPowerProfile([0.5, 2, 2])
        np.testing.assert_allclose(pp(x), tp(x))
        np.testing.assert_allclose(pp(x, dr=1), tp(x, dr=1))
        np.testing.assert_allclose(pp(x, dr=2), tp(x, dr=2))
        np.testing.assert_allclose(
            pp.params, tp.to_powerseries(order=4, sym=True).params
        )

    @pytest.mark.unit
    def test_PowerSeriesProfile_even_sym(self):
        """Test that even symmetry is enforced properly in PowerSeriesProfile."""
        pp = PowerSeriesProfile(params=np.array([4, 0, -2, 0, 3, 0, -1]), sym="auto")
        assert pp.sym == "even"  # auto symmetry detects it is even
        sp = pp.to_spline()

        pp_o = sp.to_powerseries(sym="auto")
        assert not pp_o.sym  # default conversion from spline is not symmetric

        pp_e = sp.to_powerseries(sym="even")
        assert pp_e.sym == "even"  # check that even symmetry is enforced
        # check that this matches the original parameters
        np.testing.assert_allclose(pp.params, pp_e.params, rtol=1e-3)

    @pytest.mark.unit
    def test_SplineProfile_methods(self):
        """Test that all methods of SplineProfile work as intended without errors."""
        pp = PowerSeriesProfile(
            modes=np.array([0, 2, 4]), params=np.array([1, -2, 1]), sym=False
        )  # base profile to work off of
        knots = np.linspace(0, 1.0, 40, endpoint=True)
        x = np.linspace(0, 0.99, 100)

        method = "nearest"
        sp = pp.to_spline(knots=knots, method=method)
        # should be exactly same if evaluated at knots + eps
        np.testing.assert_allclose(pp(knots), sp(knots + 0.4 * (x[1] - x[0])))

        method = "linear"
        sp = pp.to_spline(knots=knots, method=method)
        # should match linear interpolation
        scipy_interp = interp1d(x=knots, y=pp(knots))
        np.testing.assert_allclose(scipy_interp(x), sp(x))

        method = "cubic"
        sp = pp.to_spline(knots=knots, method=method)
        np.testing.assert_allclose(pp(x), sp(x), rtol=1e-5, atol=1e-3)

        method = "cubic2"
        sp = pp.to_spline(knots=knots, method=method)
        np.testing.assert_allclose(pp(x), sp(x), rtol=1e-5, atol=1e-3)

        method = "catmull-rom"
        sp = pp.to_spline(knots=knots, method=method)
        np.testing.assert_allclose(pp(x), sp(x), rtol=1e-5, atol=1e-3)

        method = "cardinal"
        sp = pp.to_spline(knots=knots, method=method)
        np.testing.assert_allclose(pp(x), sp(x), rtol=1e-5, atol=1e-3)

        method = "monotonic"
        sp = pp.to_spline(knots=knots, method=method)
        np.testing.assert_allclose(pp(x), sp(x), rtol=1e-5, atol=1e-3)

        method = "monotonic-0"
        sp = pp.to_spline(knots=knots, method=method)
        np.testing.assert_allclose(pp(x), sp(x), rtol=1e-5, atol=5e-3)

        # test monotonic splines preserve monotonicity
        f = np.heaviside(knots - 0.5, 0) + 0.1 * knots
        spm = SplineProfile(values=f, knots=knots, method="monotonic")
        spm0 = SplineProfile(values=f, knots=knots, method="monotonic-0")
        spc = SplineProfile(values=f, knots=knots, method="cubic")

        dfc = spc(x, dr=1)
        dfm = spm(x, dr=1)
        dfm0 = spm0(x, dr=1)
        assert dfc.min() < 0  # cubic interpolation undershoots, giving negative slope
        assert dfm.min() > 0  # monotonic interpolation doesn't
        assert dfm0.min() >= 0  # monotonic-0 doesn't overshoot either

    @pytest.mark.unit
    def test_repr(self):
        """Test string representation of profile classes."""
        pp = PowerSeriesProfile(modes=np.array([0, 2, 4]), params=np.array([1, -2, 1]))
        tp = TwoPowerProfile([1, 2, 1])
        sp = pp.to_spline()
        mp = pp.to_mtanh(order=4, ftol=1e-4, xtol=1e-4)
        zp = pp.to_fourierzernike()
        assert "PowerSeriesProfile" in str(pp)
        assert "TwoPowerProfile" in str(tp)
        assert "SplineProfile" in str(sp)
        assert "MTanhProfile" in str(mp)
        assert "FourierZernikeProfile" in str(zp)
        assert "SumProfile" in str(pp + zp)
        assert "ProductProfile" in str(pp * zp)
        assert "ScaledProfile" in str(2 * zp)

    @pytest.mark.unit
    def test_get_set(self):
        """Test getting/setting of profile attributes."""
        pp = PowerSeriesProfile(
            modes=np.array([0, 2, 4]), params=np.array([1, -2, 1]), sym=False
        )

        assert pp.get_params(2) == -2
        assert pp.basis.get_idx(4) == 4
        pp.set_params(3, 22)

        assert pp.get_params(3) == 22
        pp.change_resolution(L=2)
        assert pp.params.size == 3

        sp = pp.to_spline()
        sp.params = sp.params + 1
        np.testing.assert_allclose(sp.params, 1 + pp(sp._knots))

        tp = TwoPowerProfile([1, 2, 3])
        np.testing.assert_allclose(tp.params, [1, 2, 3])
        tp.params = [1 / 2, 3 / 2, 4 / 3]
        np.testing.assert_allclose(tp.params, [1 / 2, 3 / 2, 4 / 3])

        zp = FourierZernikeProfile([1 - 1 / 3 - 1 / 6, -1 / 2, 1 / 6])
        assert zp.get_params(2, 0, 0) == -1 / 2
        zp.set_params(2, 0, 0, 1 / 4)
        assert zp.get_params(2, 0, 0) == 1 / 4
        zp.change_resolution(L=0)
        assert len(zp.params) == 1

    @pytest.mark.unit
    def test_auto_sym(self):
        """Test that even parity is enforced automatically."""
        pp = PowerSeriesProfile(
            modes=np.array([0, 1, 2, 4]), params=np.array([1, 0, -2, 1]), sym="auto"
        )
        assert pp.sym == "even"
        assert pp.basis.num_modes == 3
        pp = PowerSeriesProfile(
            modes=np.array([0, 1, 2, 4]), params=np.array([1, 1, -2, 1]), sym="auto"
        )
        assert pp.sym is False
        assert pp.basis.num_modes == 5

    @pytest.mark.unit
    def test_sum_profiles(self):
        """Test adding two profiles together."""
        pp = PowerSeriesProfile(
            modes=np.array([0, 1, 2, 4]), params=np.array([1, 0, -2, 1]), sym="auto"
        )
        sp = pp.to_spline()
        zp = -pp.to_fourierzernike()

        f = pp + sp - zp
        x = np.linspace(0, 1, 50)
        np.testing.assert_allclose(f(x), 3 * (pp(x)), atol=1e-3)

        params = f.params
        assert params.size == len(sp.params) + len(pp.params) + len(zp.params) + 1
        params = f._parse_params(params)
        assert all(params[0] == pp.params)
        assert all(params[1] == sp.params)
        # offset by 1 because of two - signs
        assert all(params[2][1:] == zp.params)
        assert params[2][0] == -1

        f.params = (pp.params, 2 * sp.params, zp.params)
        np.testing.assert_allclose(f(x), 4 * (pp(x)), atol=1e-3)

    @pytest.mark.unit
    def test_product_profiles(self):
        """Test multiplying two profiles together."""
        pp = PowerSeriesProfile(
            modes=np.array([0, 1, 2, 4]), params=np.array([1, 0, -2, 1]), sym="auto"
        )
        sp = pp.to_spline()
        zp = pp.to_fourierzernike()

        f = pp * sp * zp
        x = np.linspace(0, 1, 50)
        np.testing.assert_allclose(f(x), pp(x) ** 3, atol=1e-3)

        params = f.params
        assert params.size == len(sp.params) + len(pp.params) + len(zp.params)
        params = f._parse_params(params)
        assert all(params[0] == pp.params)
        assert all(params[1] == sp.params)
        assert all(params[2] == zp.params)

        f.params = (pp.params, 2 * sp.params, zp.params)
        np.testing.assert_allclose(f(x), 2 * pp(x) ** 3, atol=1e-3)

    @pytest.mark.unit
    def test_product_profiles_derivative(self):
        """Test that product profiles computes the derivative correctly."""
        p1 = PowerSeriesProfile(
            modes=np.array([0, 1, 2, 4]), params=np.array([1, 3, -2, 4]), sym="auto"
        )
        p2 = PowerSeriesProfile(
            modes=np.array([0, 1, 2, 3]),
            params=np.array([2.02, 4.93, 0.22, 0.46]),
            sym="auto",
        )
        p3 = PowerSeriesProfile(
            modes=np.array([0, 1, 3, 4]),
            params=np.array([1.79, 3.19, 1.82, 2.07]),
            sym="auto",
        )

        f = p1 * p2 * p3
        x = np.linspace(0, 1, 50)

        # Below is a simpler method to compute first derivative of product series
        # than the more general combinatorics algorithm in implementation.
        # Analytic formula derived from logarithmic differentiation.
        _sum = 0
        _sum_r = 0
        for profile in f._profiles:
            result = profile(x)
            result_r = profile(x, dr=1)
            result_rr = profile(x, dr=2)
            _sum += result_r / result
            _sum_r += result_rr / result - (result_r / result) ** 2
        f_r = f(x) * _sum
        f_rr = f_r * _sum + f(x) * _sum_r

        np.testing.assert_allclose(f_r, f(x, dr=1))
        np.testing.assert_allclose(f_rr, f(x, dr=2))

    @pytest.mark.unit
    def test_scaled_profiles(self):
        """Test scaling profiles by a constant."""
        pp = PowerSeriesProfile(
            modes=np.array([0, 1, 2, 4]), params=np.array([1, 0, -2, 1]), sym="auto"
        )

        f = 3 * pp
        x = np.linspace(0, 1, 50)
        np.testing.assert_allclose(f(x), 3 * (pp(x)), atol=1e-3)

        params = f.params
        assert params[0] == 3
        assert all(params[1:] == pp.params)

        f.params = 2
        np.testing.assert_allclose(f(x), 2 * (pp(x)), atol=1e-3)

        f.params = 4 * pp.params

        params = f.params
        assert params.size == len(pp.params) + 1
        assert params[0] == 2
        np.testing.assert_allclose(params[1:], [4, -8, 4])
        np.testing.assert_allclose(pp.params, [1, -2, 1])
        np.testing.assert_allclose(f(x), 8 * (pp(x)), atol=1e-3)

    @pytest.mark.unit
    def test_profile_errors(self):
        """Test error checking when creating and working with profiles."""
        pp = PowerSeriesProfile(
            modes=np.array([0, 1, 2, 4]), params=np.array([1, 0, -2, 1]), sym="auto"
        )
        sp = pp.to_spline()
        zp = pp.to_fourierzernike()
        mp = pp.to_mtanh(order=4, ftol=1e-4, xtol=1e-4)
        tp = TwoPowerProfile(params=np.array([0.5, 2, 1.5]))
        grid = LinearGrid(L=9)

        with pytest.raises(ValueError):
            zp.params = 4
        with pytest.raises(ValueError):
            mp.params = np.arange(4)
        with pytest.raises(ValueError):
            sp.params = np.arange(4)
        with pytest.raises(ValueError):
            pp.params = np.arange(8)
        with pytest.raises(NotImplementedError):
            a = sp + 4
        with pytest.raises(NotImplementedError):
            a = sp * [1, 2, 3]
        with pytest.raises(ValueError):
            a = sp + pp
            a.params = pp.params
        with pytest.raises(ValueError):
            a = sp * pp
            a.params = sp.params
        with pytest.raises(ValueError):
            _ = TwoPowerProfile([1, 2, 3, 4])
        with pytest.raises(NotImplementedError):
            tp.compute(grid, dr=3)
        with pytest.raises(NotImplementedError):
            mp.compute(grid, dr=3)

    @pytest.mark.unit
    def test_default_profiles(self):
        """Test that default profiles are just zeros."""
        pp = PowerSeriesProfile()
        tp = TwoPowerProfile()
        sp = SplineProfile()
        mp = MTanhProfile()
        zp = FourierZernikeProfile()

        x = np.linspace(0, 1, 10)
        np.testing.assert_allclose(pp(x), 0)
        np.testing.assert_allclose(tp(x), 0)
        np.testing.assert_allclose(sp(x), 0)
        np.testing.assert_allclose(mp(x), 0)
        np.testing.assert_allclose(zp(x), 0)

    @pytest.mark.regression
    def test_solve_with_combined(self):
        """Make sure combined profiles work correctly for solving equilibrium.

        Test for GH issue #347.
        """
        ne = PowerSeriesProfile(3.0e19 * np.array([1, -1]), modes=[0, 10])
        Te = PowerSeriesProfile(2.0e3 * np.array([1, -1]), modes=[0, 2])
        Ti = Te
        pressure = elementary_charge * (ne * Te + ne * Ti)

        LM_resolution = 6
        eq1 = Equilibrium(
            pressure=pressure,
            iota=PowerSeriesProfile([1.61]),
            Psi=np.pi,  # so B ~ 1 T
            NFP=1,
            L=LM_resolution,
            M=LM_resolution,
            N=0,
            L_grid=2 * LM_resolution,
            M_grid=2 * LM_resolution,
            N_grid=0,
            sym=True,
        )
        eq1.solve(
            constraints=get_fixed_boundary_constraints(eq=eq1),
            objective=ObjectiveFunction(objectives=ForceBalance(eq=eq1)),
            maxiter=5,
        )
        eq2 = Equilibrium(
            electron_temperature=Te,
            electron_density=ne,
            atomic_number=1,
            ion_temperature=Ti,
            iota=PowerSeriesProfile([1.61]),
            Psi=np.pi,  # so B ~ 1 T
            NFP=1,
            L=LM_resolution,
            M=LM_resolution,
            N=0,
            L_grid=2 * LM_resolution,
            M_grid=2 * LM_resolution,
            N_grid=0,
            sym=True,
        )
        eq2.solve(
            constraints=get_fixed_boundary_constraints(eq=eq2),
            objective=ObjectiveFunction(objectives=ForceBalance(eq=eq2)),
            maxiter=5,
        )
        np.testing.assert_allclose(eq1.R_lmn, eq2.R_lmn, atol=1e-14)
        np.testing.assert_allclose(eq1.Z_lmn, eq2.Z_lmn, atol=1e-14)
        np.testing.assert_allclose(eq1.L_lmn, eq2.L_lmn, atol=1e-14)

    @pytest.mark.unit
    def test_kinetic_pressure(self):
        """Test that both ways of computing pressure are equivalent."""
        ne = PowerSeriesProfile(3.0e19 * np.array([1, -1]), modes=[0, 10])
        Te = PowerSeriesProfile(2.0e3 * np.array([1, -1]), modes=[0, 2])
        Ti = Te
        pressure = elementary_charge * (ne * Te + ne * Ti)

        LM_resolution = 6
        eq1 = Equilibrium(
            pressure=pressure,
            iota=PowerSeriesProfile([1.61]),
            Psi=np.pi,  # so B ~ 1 T
            NFP=1,
            L=LM_resolution,
            M=LM_resolution,
            N=0,
            L_grid=2 * LM_resolution,
            M_grid=2 * LM_resolution,
            N_grid=0,
            sym=True,
        )
        eq2 = Equilibrium(
            electron_temperature=Te,
            electron_density=ne,
            iota=PowerSeriesProfile([1.61]),
            Psi=np.pi,  # so B ~ 1 T
            NFP=1,
            L=LM_resolution,
            M=LM_resolution,
            N=0,
            L_grid=2 * LM_resolution,
            M_grid=2 * LM_resolution,
            N_grid=0,
            sym=True,
        )
        grid = LinearGrid(L=20)
        data1 = eq1.compute(["p", "p_r"], grid=grid)
        data2 = eq2.compute(["p", "p_r"], grid=grid)

        assert np.all(np.isnan(data1["ne"]))
        assert np.all(np.isnan(data1["Te"]))
        assert np.all(np.isnan(data1["Ti"]))
        assert np.all(np.isnan(data1["Zeff"]))
        assert np.all(np.isnan(data1["ne_r"]))
        assert np.all(np.isnan(data1["Te_r"]))
        assert np.all(np.isnan(data1["Ti_r"]))
        assert np.all(np.isnan(data1["Zeff_r"]))
        assert np.all(data2["Te"] == data2["Ti"])
        assert np.all(data2["Te_r"] == data2["Ti_r"])
        np.testing.assert_allclose(data1["p"], data2["p"])
        np.testing.assert_allclose(data1["p_r"], data2["p_r"])

    @pytest.mark.unit
    def test_hermite_spline_solve(self):
        """Test that spline with double number of parameters is optimized."""
        eq = get("DSHAPE")
        rho = np.linspace(0, 1.0, 20, endpoint=True)
        eq.pressure = HermiteSplineProfile(
            eq.pressure(rho), eq.pressure(rho, dr=1), rho
        )
        eq.solve()
        assert eq.is_nested()



"""Utilities for quadratures."""

from orthax.legendre import legder, legval

from desc.backend import eigh_tridiagonal, jnp, put
from desc.utils import errorif


def bijection_to_disc(x, a, b):
    """[a, b] ∋ x ↦ y ∈ [−1, 1]."""
    y = 2.0 * (x - a) / (b - a) - 1.0
    return y


def bijection_from_disc(x, a, b):
    """[−1, 1] ∋ x ↦ y ∈ [a, b]."""
    y = 0.5 * (b - a) * (x + 1.0) + a
    return y


def grad_bijection_from_disc(a, b):
    """Gradient wrt ``x`` of ``bijection_from_disc``."""
    dy_dx = 0.5 * (b - a)
    return dy_dx


def automorphism_arcsin(x):
    """[-1, 1] ∋ x ↦ y ∈ [−1, 1].

    This map decreases node density near the boundary by the asymptotic factor
    √(1−x²) and adds a 1/√(1−x²) factor to the integrand.

    Parameters
    ----------
    x : jnp.ndarray
        Points to transform.

    Returns
    -------
    y : jnp.ndarray
        Transformed points.

    """
    y = 2.0 * jnp.arcsin(x) / jnp.pi
    return y


def grad_automorphism_arcsin(x):
    """Gradient of arcsin automorphism."""
    dy_dx = 2.0 / (jnp.sqrt(1.0 - x**2) * jnp.pi)
    return dy_dx


grad_automorphism_arcsin.__doc__ += "\n" + automorphism_arcsin.__doc__


def automorphism_sin(x, s=0, m=10):
    """[-1, 1] ∋ x ↦ y ∈ [−1, 1].

    This map increases node density near the boundary by the asymptotic factor
    1/√(1−x²) and adds a √(1−x²) factor to the integrand.

    Parameters
    ----------
    x : jnp.ndarray
        Points to transform.
    s : float
        Strength of derivative suppression, s ∈ [0, 1].
    m : float
        Number of machine epsilons used for floating point error buffer.

    Returns
    -------
    y : jnp.ndarray
        Transformed points.

    """
    errorif(not (0 <= s <= 1))
    # s = 0 -> derivative vanishes like cosine.
    # s = 1 -> derivative vanishes like cosine^k.
    y0 = jnp.sin(0.5 * jnp.pi * x)
    y1 = x + jnp.sin(jnp.pi * x) / jnp.pi  # k = 2
    y = (1 - s) * y0 + s * y1
    # y is an expansion, so y(x) > x near x ∈ {−1, 1} and there is a tendency
    # for floating point error to overshoot the true value.
    eps = m * jnp.finfo(jnp.array(1.0).dtype).eps
    return jnp.clip(y, -1 + eps, 1 - eps)


def grad_automorphism_sin(x, s=0):
    """Gradient of sin automorphism."""
    dy0_dx = 0.5 * jnp.pi * jnp.cos(0.5 * jnp.pi * x)
    dy1_dx = 1.0 + jnp.cos(jnp.pi * x)
    dy_dx = (1 - s) * dy0_dx + s * dy1_dx
    return dy_dx


grad_automorphism_sin.__doc__ += "\n" + automorphism_sin.__doc__


def tanh_sinh(deg, m=10):
    """Tanh-Sinh quadrature.

    Returns quadrature points xₖ and weights wₖ for the approximate evaluation
    of the integral ∫₋₁¹ f(x) dx ≈ ∑ₖ wₖ f(xₖ).

    Parameters
    ----------
    deg : int
        Number of quadrature points.
    m : float
        Number of machine epsilons used for floating point error buffer. Larger
        implies less floating point error, but increases the minimum achievable
        quadrature error.

    Returns
    -------
    x, w : (jnp.ndarray, jnp.ndarray)
        Shape (deg, ).
        Quadrature points and weights.

    """
    # buffer to avoid numerical instability
    x_max = jnp.array(1.0)
    x_max = x_max - m * jnp.finfo(x_max.dtype).eps
    t_max = jnp.arcsinh(2 * jnp.arctanh(x_max) / jnp.pi)
    # maximal-spacing scheme, doi.org/10.48550/arXiv.2007.15057
    t = jnp.linspace(-t_max, t_max, deg)
    dt = 2 * t_max / (deg - 1)
    arg = 0.5 * jnp.pi * jnp.sinh(t)
    x = jnp.tanh(arg)  # x = g(t)
    w = 0.5 * jnp.pi * jnp.cosh(t) / jnp.cosh(arg) ** 2 * dt  # w = (dg/dt) dt
    return x, w


def leggauss_lob(deg, interior_only=False):
    """Lobatto-Gauss-Legendre quadrature.

    Returns quadrature points xₖ and weights wₖ for the approximate evaluation
    of the integral ∫₋₁¹ f(x) dx ≈ ∑ₖ wₖ f(xₖ).

    Parameters
    ----------
    deg : int
        Number of quadrature points.
    interior_only : bool
        Whether to exclude the points and weights at -1 and 1;
        useful if f(-1) = f(1) = 0. If true, then ``deg`` points are still
        returned; these are the interior points for lobatto quadrature of ``deg+2``.

    Returns
    -------
    x, w : (jnp.ndarray, jnp.ndarray)
        Shape (deg, ).
        Quadrature points and weights.

    """
    N = deg + 2 * bool(interior_only)
    errorif(N < 2)

    # Golub-Welsh algorithm
    n = jnp.arange(2, N - 1)
    x = eigh_tridiagonal(
        jnp.zeros(N - 2),
        jnp.sqrt((n**2 - 1) / (4 * n**2 - 1)),
        eigvals_only=True,
    )
    c0 = put(jnp.zeros(N), -1, 1)

    # improve (single multiplicity) roots by one application of Newton
    c = legder(c0)
    dy = legval(x=x, c=c)
    df = legval(x=x, c=legder(c))
    x -= dy / df

    w = 2 / (N * (N - 1) * legval(x=x, c=c0) ** 2)

    if not interior_only:
        x = jnp.hstack([-1.0, x, 1.0])
        w_end = 2 / (deg * (deg - 1))
        w = jnp.hstack([w_end, w, w_end])

    assert x.size == w.size == deg
    return x, w


def uniform(deg):
    """Uniform quadrature that is Gauss-Chebyshev in transformed variable.

    Returns quadrature points xₖ and weights wₖ for the approximate evaluation
    of the integral ∫₋₁¹ f(x) dx ≈ ∑ₖ wₖ f(xₖ).

    Parameters
    ----------
    deg : int
        Number of quadrature points.

    Returns
    -------
    x, w : (jnp.ndarray, jnp.ndarray)
        Shape (deg, ).
        Quadrature points and weights.

    """
    # Define x = 2/π arcsin y and g : y ↦ f(x(y)).
    #   ∫₋₁¹ f(x) dx = 2/π ∫₋₁¹ (1−y²)⁻⁰ᐧ⁵ g(y) dy
    # ∑ₖ wₖ f(x(yₖ)) = 2/π ∑ₖ ωₖ g(yₖ)
    # Given roots yₖ of Chebyshev polynomial, x(yₖ) below is uniform in (-1, 1).
    x = jnp.arange(-deg + 1, deg + 1, 2) / deg
    w = 2 / deg * jnp.ones(deg)
    return x, w


def chebgauss2(deg):
    """Gauss-Chebyshev quadrature of the second kind with implicit weighting.

    Returns quadrature points xₖ and weights wₖ for the approximate evaluation
    of the integral ∫₋₁¹ f(x) dx ≈ ∑ₖ wₖ f(xₖ) where f(x) = g(x) √(1−x²).

    Parameters
    ----------
    deg : int
        Number of quadrature points.

    Returns
    -------
    x, w : (jnp.ndarray, jnp.ndarray)
        Shape (deg, ).
        Quadrature points and weights.

    """
    # Adapted from
    # github.com/scipy/scipy/blob/v1.14.1/scipy/special/_orthogonal.py#L1803-L1851.
    t = jnp.arange(deg, 0, -1) * jnp.pi / (deg + 1)
    x = jnp.cos(t)
    w = jnp.pi * jnp.abs(jnp.sin(t)) / (deg + 1)
    return x, w


def get_quadrature(quad, automorphism):
    """Apply automorphism to given quadrature.

    Parameters
    ----------
    quad : (jnp.ndarray, jnp.ndarray)
        Quadrature points xₖ and weights wₖ for the approximate evaluation of
        the integral ∫₋₁¹ g(x) dx = ∑ₖ wₖ g(xₖ).
    automorphism : (Callable, Callable) or None
        The first callable should be an automorphism of the real interval [-1, 1].
        The second callable should be the derivative of the first. This map defines
        a change of variable for the bounce integral. The choice made for the
        automorphism will affect the performance of the quadrature method.

    Returns
    -------
    x, w : (jnp.ndarray, jnp.ndarray)
        Quadrature points and weights.

    """
    x, w = quad
    assert x.ndim == w.ndim == 1
    if automorphism is not None:
        auto, grad_auto = automorphism
        w = w * grad_auto(x)
        # Recall bijection_from_disc(auto(x), ζ₁, ζ₂) = ζ.
        x = auto(x)
    return x, w



"""Functions for converting between coordinate systems."""

import functools

from desc.backend import jnp

from ..utils import safenorm, safenormalize


def reflection_matrix(normal):
    """Matrix to reflect points across plane through origin with specified normal.

    Parameters
    ----------
    normal : array-like, shape(3,)
        Vector normal to plane of reflection, in cartesian (X,Y,Z) coordinates

    Returns
    -------
    flip : ndarray, shape(3,3)
        Matrix to flip points in cartesian (X,Y,Z) coordinates
    """
    normal = jnp.asarray(normal)
    R = jnp.eye(3) - 2 * jnp.outer(normal, normal) / jnp.inner(normal, normal)
    return R


def rotation_matrix(axis, angle=None):
    """Matrix to rotate points about axis by given angle.

    Parameters
    ----------
    axis : array-like, shape(3,)
        Axis of rotation, in cartesian (X,Y,Z) coordinates
    angle : float or None
        Angle to rotate by, in radians. If None, use norm of axis vector.

    Returns
    -------
    rotmat : ndarray, shape(3,3)
        Matrix to rotate points in cartesian (X,Y,Z) coordinates.

    """
    axis = jnp.asarray(axis)
    norm = safenorm(axis)
    axis = safenormalize(axis)
    if angle is None:
        angle = norm
    eps = 1e2 * jnp.finfo(axis.dtype).eps
    R1 = jnp.cos(angle) * jnp.eye(3)
    R2 = jnp.sin(angle) * jnp.cross(axis, jnp.identity(axis.shape[0]) * -1)
    R3 = (1 - jnp.cos(angle)) * jnp.outer(axis, axis)
    return jnp.where(norm < eps, jnp.eye(3), R1 + R2 + R3)  # if axis=0, no rotation


def xyz2rpz(pts):
    """Transform points from cartesian (X,Y,Z) to polar (R,phi,Z) form.

    Parameters
    ----------
    pts : ndarray, shape(...,3)
        points in cartesian (X,Y,Z) coordinates

    Returns
    -------
    pts : ndarray, shape(...,3)
        points in polar (R,phi,Z) coordinates

    """
    x, y, z = pts.T
    r = jnp.sqrt(x**2 + y**2)
    p = jnp.arctan2(y, x)
    return jnp.array([r, p, z]).T


def rpz2xyz(pts):
    """Transform points from polar (R,phi,Z) to cartesian (X,Y,Z) form.

    Parameters
    ----------
    pts : ndarray, shape(...,3)
        points in polar (R,phi,Z) coordinates

    Returns
    -------
    pts : ndarray, shape(...,3)
        points in cartesian (X,Y,Z) coordinates

    """
    r, p, z = pts.T
    x = r * jnp.cos(p)
    y = r * jnp.sin(p)
    return jnp.array([x, y, z]).T


def xyz2rpz_vec(vec, x=None, y=None, phi=None):
    """Transform vectors from cartesian (X,Y,Z) to polar (R,phi,Z) form.

    Parameters
    ----------
    vec : ndarray, shape(...,3)
        vectors, in cartesian (X,Y,Z) form
    x, y, phi : ndarray, shape(...,)
        anchor points for vectors. Either x and y, or phi must be supplied

    Returns
    -------
    vec : ndarray, shape(...,3)
        vectors, in polar (R,phi,Z) form

    """
    if x is not None and y is not None:
        phi = jnp.arctan2(y, x)

    @functools.partial(jnp.vectorize, signature="(3),()->(3)")
    def inner(vec, phi):
        rot = jnp.array(
            [
                [jnp.cos(phi), -jnp.sin(phi), jnp.zeros_like(phi)],
                [jnp.sin(phi), jnp.cos(phi), jnp.zeros_like(phi)],
                [jnp.zeros_like(phi), jnp.zeros_like(phi), jnp.ones_like(phi)],
            ]
        )
        rot = rot.T
        polar = jnp.matmul(rot, vec)
        return polar

    return inner(vec, phi)


def rpz2xyz_vec(vec, x=None, y=None, phi=None):
    """Transform vectors from polar (R,phi,Z) to cartesian (X,Y,Z) form.

    Parameters
    ----------
    vec : ndarray, shape(n,3)
        vectors, in polar (R,phi,Z) form
    x, y, phi : ndarray, shape(n,)
        anchor points for vectors. Either x and y, or phi must be supplied

    Returns
    -------
    vec : ndarray, shape(n,3)
        vectors, in cartesian (X,Y,Z) form

    """
    if x is not None and y is not None:
        phi = jnp.arctan2(y, x)

    @functools.partial(jnp.vectorize, signature="(3),()->(3)")
    def inner(vec, phi):
        rot = jnp.array(
            [
                [jnp.cos(phi), jnp.sin(phi), jnp.zeros_like(phi)],
                [-jnp.sin(phi), jnp.cos(phi), jnp.zeros_like(phi)],
                [jnp.zeros_like(phi), jnp.zeros_like(phi), jnp.ones_like(phi)],
            ]
        )
        rot = rot.T
        cart = jnp.matmul(rot, vec)
        return cart

    return inner(vec, phi)



"""Example script for recreating the precise QA configuration of Landreman and Paul."""

from desc import set_device

# need to do this before importing other DESC stuff so JAX initializes properly
set_device("gpu")

import numpy as np

from desc.continuation import solve_continuation_automatic
from desc.equilibrium import EquilibriaFamily, Equilibrium
from desc.geometry import FourierRZToroidalSurface
from desc.grid import LinearGrid
from desc.objectives import (
    AspectRatio,
    FixBoundaryR,
    FixBoundaryZ,
    FixCurrent,
    FixPressure,
    FixPsi,
    ForceBalance,
    ObjectiveFunction,
    QuasisymmetryTwoTerm,
    RotationalTransform,
)
from desc.optimize import Optimizer

# create initial surface. Aspect ratio ~ 6, circular cross section with slight
# axis torsion to make it nonplanar
surf = FourierRZToroidalSurface(
    R_lmn=[1, 0.166, 0.1],
    Z_lmn=[-0.166, -0.1],
    modes_R=[[0, 0], [1, 0], [0, 1]],
    modes_Z=[[-1, 0], [0, -1]],
    NFP=2,
)
# create initial equilibrium. Psi chosen to give B ~ 1 T. Could also give profiles here,
# default is zero pressure and zero current
eq = Equilibrium(M=8, N=8, Psi=0.087, surface=surf)
# this is usually all you need to solve a fixed boundary equilibrium
eq = solve_continuation_automatic(eq, objective="force", bdry_step=0.5, verbose=3)[-1]
# it will be helpful to store intermediate results
eqfam = EquilibriaFamily(eq)

# create grid where we want to minimize QS error. Here we do it on 3 surfaces
grid = LinearGrid(M=eq.M, N=eq.N, NFP=eq.NFP, rho=np.array([0.6, 0.8, 1.0]), sym=True)

# optimize in steps
for k in range(1, eq.M + 1):
    print("\n==================================")
    print("Optimizing boundary modes M,N <= {}".format(k))
    print("====================================")
    objective = ObjectiveFunction(
        (
            # pass in the grid we defined, and don't forget the target helicity!
            QuasisymmetryTwoTerm(
                eq=eqfam[-1], helicity=(1, 0), grid=grid, normalize=False
            ),
            AspectRatio(eq=eqfam[-1], target=6, weight=1e1, normalize=False),
            # this targets a profile pointwise, which is ok because we expect it to be
            # fairly flat
            RotationalTransform(eq=eqfam[-1], target=0.42, weight=10, normalize=False),
            # we could optionally set normalize=True which would compute things in
            # normalized/dimensionless units, effectively changing the weights
        ),
    )
    # as opposed to SIMSOPT and STELLOPT where variables are assumed fixed, in DESC
    # we assume variables are free. Here we decide which ones to fix, starting with
    # the major radius (R mode = [0,0,0]) and all modes with m,n > k
    R_modes = np.vstack(
        (
            [0, 0, 0],
            eq.surface.R_basis.modes[
                np.max(np.abs(eq.surface.R_basis.modes), 1) > k, :
            ],
        )
    )
    Z_modes = eq.surface.Z_basis.modes[
        np.max(np.abs(eq.surface.Z_basis.modes), 1) > k, :
    ]
    # next we create the constraints, using the mode number arrays just created
    # if we didn't pass those in, it would fix all the modes (like for the profiles)
    constraints = (
        ForceBalance(eq=eqfam[-1]),  # J x B - grad(p) = 0
        FixBoundaryR(eq=eqfam[-1], modes=R_modes),
        FixBoundaryZ(eq=eqfam[-1], modes=Z_modes),
        FixPressure(eq=eqfam[-1]),
        FixCurrent(eq=eqfam[-1]),
        FixPsi(eq=eqfam[-1]),
    )
    # this is the default optimizer, which re-solves the equilibrium at each step
    optimizer = Optimizer("proximal-lsq-exact")
    # we get a new equilibrium by optimizing the old one and passing copy=True.
    # otherwise, we could modify the original equilibrium in place
    eq_new, out = eqfam[-1].optimize(
        objective=objective,
        constraints=constraints,
        optimizer=optimizer,
        maxiter=50,
        verbose=3,
        copy=True,
        options={
            # sometimes the default initial trust radius is too big, allowing the
            # optimizer to take too large a step in a bad direction. If this happens,
            # we can manually specify a smaller starting radius.
            "initial_trust_radius": 0.5,
        },
    )
    # add our new equilibrium to the family
    eqfam.append(eq_new)

# save all the steps of the optimization for later analysis
eqfam.save("precise_QA_output.h5")



"""Tests for reading/writing input/output, both ascii and binary."""

import os
import pathlib
import shutil

import h5py
import numpy as np
import pytest

import desc.examples
from desc.basis import FourierZernikeBasis
from desc.equilibrium import Equilibrium
from desc.grid import LinearGrid
from desc.io import InputReader, hdf5Reader, hdf5Writer, load
from desc.io.ascii_io import read_ascii, write_ascii
from desc.magnetic_fields import (
    OmnigenousField,
    SplineMagneticField,
    ToroidalMagneticField,
)
from desc.transform import Transform
from desc.utils import equals


@pytest.mark.unit
def test_vmec_input(tmpdir_factory):
    """Test converting VMEC to DESC input file."""
    # input.DSHAPE has multi-line inputs and
    # a duplicate AXIS input line as well, so
    # exercises full VMEC capability of input reader
    input_path = "./tests/inputs/input.DSHAPE"
    tmpdir = tmpdir_factory.mktemp("desc_inputs")
    tmp_path = tmpdir.join("input.DSHAPE")
    shutil.copyfile(input_path, tmp_path)
    with pytest.warns(UserWarning):
        ir = InputReader(cl_args=[str(tmp_path)])
    vmec_inputs = ir.inputs
    # ir makes a VMEC file automatically
    path_converted_file = tmpdir.join("input.DSHAPE_desc")
    # also test making a DESC file from the ir.inputs manually
    path = tmpdir.join("desc_from_vmec")
    ir.write_desc_input(path, ir.inputs)
    ir2 = InputReader(cl_args=[str(path)])
    desc_inputs = ir2.inputs
    for d, v in zip(desc_inputs, vmec_inputs):
        d.pop("output_path")
        v.pop("output_path")
    assert all([equals(in1, in2) for in1, in2 in zip(vmec_inputs, desc_inputs)])

    correct_file_path = ".//tests//inputs//input.DSHAPE_desc"

    # check DESC input file matches known correct one line-by-line
    with open(correct_file_path) as f:
        lines_correct = f.readlines()
    with open(path) as f:
        lines_direct = f.readlines()
    with open(path_converted_file) as f:
        lines_converted = f.readlines()
    # skip first 4 lines as they have date and pwd info
    for line1, line2 in zip(lines_correct[4:], lines_converted[4:]):
        assert line1.strip() == line2.strip()
    for line1, line2 in zip(lines_correct[4:], lines_direct):
        assert line1.strip() == line2.strip()


@pytest.mark.unit
def test_write_desc_input_Nones(tmpdir_factory):
    """Test converting writing DESC input file when an input tol is None."""
    # tests how None is handled as a passed-in input to the input writer
    # if None is passed for one of the elements of an input item
    # such as ftol, gtol, xtol or nfev,
    # then that will not be written
    # for example if inputs['ftol'] = [1e-2,None,1e-3]
    # the written file will have ftol = 0.01,0.001
    # and the None will have not been written

    # if only Nones are passed for just one of the tolerances, only that one will not
    # be written. gtol is used here as that test

    input_path = "./tests/inputs/DSHAPE"
    tmpdir = tmpdir_factory.mktemp("desc_inputs")
    tmp_path = tmpdir.join("DSHAPE")
    shutil.copyfile(input_path, tmp_path)
    ir = InputReader(cl_args=[str(tmp_path)])

    ftols_input_with_none = [1e-2, None, 1e-3]
    for i, inp in enumerate(ir.inputs):
        inp["ftol"] = ftols_input_with_none[i]
        inp["gtol"] = None

    path = tmpdir.join("desc_with_None")
    ir.write_desc_input(path, ir.inputs)
    ir2 = InputReader(cl_args=[str(path)])
    correct_ftols = [1e-2, 1e-3, 1e-3]
    for i, inp in enumerate(ir2.inputs):
        assert inp["ftol"] == correct_ftols[i]

    # now check that the written line is
    # the correct "ftol = 1e-2, 1e-3"
    # and that gtol is NOT written anywhere,
    # since only None was passed in for that input
    no_gtol = True
    with open(path) as f:
        lines = f.readlines()
    for line in lines:
        if line.find("ftol") != -1:
            # line is like "ftol = 0.01, 0.001\n"
            line = line.strip().split("=")[1]
            # now we have [" 0.01, 0.001"]
            line = line.strip().split(",")
            for i, string_num in enumerate(line):
                assert float(string_num) == correct_ftols[i]
        elif line.find("gtol") != -1:
            no_gtol = False
    assert no_gtol  # fail test if gtol line is written


@pytest.mark.unit
def test_desc_output_to_input(tmpdir_factory):
    """
    Test converting DESC output to a DESC input file.

    To do that, we convert a DESC output file to a DESC input file
    named desc_input_converted. We compare the boundary Fourier
    coefficients of this file with the true values of the Fourier
    coefficients given in desc_input_truth.
    """
    outfile_path = "./tests/inputs/LandremanPaul2022_QA_reactorScale_lowRes.h5"
    tmpdir = tmpdir_factory.mktemp("desc_inputs")
    tmp_path = tmpdir.join("input_LandremanPaul_QA")
    tmpout_path = tmpdir.join("LandremanPaul2022_QA_reactorScale_lowRes.h5")
    shutil.copyfile(outfile_path, tmpout_path)

    ir1 = InputReader()
    ir1.desc_output_to_input(str(tmp_path), str(tmpout_path))
    ir1 = InputReader(cl_args=[str(tmp_path)])
    arr1 = ir1.parse_inputs()[-1]["surface"]
    arr1 = arr1[arr1[:, 1].argsort()]
    arr1mneg = arr1[arr1[:, 1] < 0]
    arr1mpos = arr1[arr1[:, 1] >= 0]

    desc_input_truth = "./tests/inputs/LandremanPaul2022_QA_reactorScale_lowRes"
    with pytest.warns(UserWarning):
        ir2 = InputReader(cl_args=[str(desc_input_truth)])
        arr2 = ir2.parse_inputs()[-1]["surface"]
    arr2 = arr2[arr2[:, 1].argsort()]
    arr2mneg = arr2[arr2[:, 1] < 0]
    arr2mpos = arr2[arr2[:, 1] >= 0]

    np.testing.assert_allclose(
        np.minimum(
            np.linalg.norm(arr1mneg[:, 3:] - arr2mneg[:, 3:]),
            np.linalg.norm(arr1mneg[:, 3:] + arr2mneg[:, 3:]),
        ),
        0,
        atol=1e-8,
    )
    np.testing.assert_allclose(
        np.minimum(
            np.linalg.norm(arr1mpos[:, 3:] - arr2mpos[:, 3:]),
            np.linalg.norm(arr1mpos[:, 3:] + arr2mpos[:, 3:]),
        ),
        0,
        atol=1e-8,
    )

    outfile_path = "./tests/inputs/iotest_HELIOTRON.h5"
    tmpdir = tmpdir_factory.mktemp("desc_inputs")
    tmp_path = tmpdir.join("input_iotest_HELIOTRON")
    tmpout_path = tmpdir.join("iotest_HELIOTRON.h5")
    shutil.copyfile(outfile_path, tmpout_path)

    ir1 = InputReader()
    ir1.desc_output_to_input(str(tmp_path), str(tmpout_path))
    ir1 = InputReader(cl_args=[str(tmp_path)])
    arr1 = ir1.parse_inputs()[-1]["surface"]
    pres1 = ir1.parse_inputs()[-1]["pressure"]

    arr1 = arr1[arr1[:, 1].argsort()]
    arr1mneg = arr1[arr1[:, 1] < 0]
    arr1mpos = arr1[arr1[:, 1] >= 0]

    desc_input_truth = "./tests/inputs/iotest_HELIOTRON"
    ir2 = InputReader(cl_args=[str(desc_input_truth)])
    arr2 = ir2.parse_inputs()[-1]["surface"]
    pres2 = ir2.parse_inputs()[-1]["pressure"]
    arr2 = arr2[arr2[:, 1].argsort()]
    arr2mneg = arr2[arr2[:, 1] < 0]
    arr2mpos = arr2[arr2[:, 1] >= 0]

    np.testing.assert_allclose(
        np.minimum(
            np.linalg.norm(arr1mneg[:, 3:] - arr2mneg[:, 3:]),
            np.linalg.norm(arr1mneg[:, 3:] + arr2mneg[:, 3:]),
        ),
        0,
        atol=1e-8,
    )

    np.testing.assert_allclose(
        np.minimum(
            np.linalg.norm(arr1mpos[:, 3:] - arr2mpos[:, 3:]),
            np.linalg.norm(arr1mpos[:, 3:] + arr2mpos[:, 3:]),
        ),
        0,
        atol=1e-8,
    )

    np.testing.assert_allclose(pres1[pres1[:, 1] > 0], pres2[pres2[:, 1] > 0])


@pytest.mark.unit
def test_near_axis_input_files():
    """Test that DESC and VMEC input files generated by pyQSC give the same inputs."""
    vmec_path = ".//tests//inputs//input.QSC_r2_5.5_vmec"
    desc_path = ".//tests//inputs//input.QSC_r2_5.5_desc"
    with pytest.warns(UserWarning):
        inputs_vmec = InputReader(vmec_path).inputs[-1]
    inputs_desc = InputReader(desc_path).inputs[-1]
    for arg in ["sym", "NFP", "Psi", "pressure", "current", "surface", "axis"]:
        np.testing.assert_allclose(
            inputs_desc[arg], inputs_vmec[arg], rtol=1e-6, atol=1e-8
        )
    if os.path.exists(".//tests//inputs//input.QSC_r2_5.5_vmec_desc"):
        os.remove(".//tests//inputs//input.QSC_r2_5.5_vmec_desc")


@pytest.mark.unit
def test_from_input_file_equilibrium_desc_vmec_DSHAPE():
    """Test that from_input_file works for DESC input files."""
    vmec_path = ".//tests//inputs//input.DSHAPE"
    desc_path = ".//tests//inputs//input.DSHAPE_desc"
    kwargs = {"spectral_indexing": "fringe"}
    with pytest.warns(UserWarning, match="Left handed"):
        eq = Equilibrium.from_input_file(desc_path, **kwargs)
    with pytest.warns(UserWarning):
        eq_VMEC = Equilibrium.from_input_file(vmec_path, **kwargs)

    # make sure the loaded eqs are equivalent
    np.testing.assert_allclose(eq.R_lmn, eq_VMEC.R_lmn)
    np.testing.assert_allclose(eq.Z_lmn, eq_VMEC.Z_lmn)
    np.testing.assert_allclose(eq.L_lmn, eq_VMEC.L_lmn)
    np.testing.assert_allclose(eq.Rb_lmn, eq_VMEC.Rb_lmn)
    np.testing.assert_allclose(eq.Zb_lmn, eq_VMEC.Zb_lmn)
    np.testing.assert_allclose(eq.Ra_n, eq_VMEC.Ra_n)
    np.testing.assert_allclose(eq.Za_n, eq_VMEC.Za_n)
    np.testing.assert_allclose(eq.Psi, eq_VMEC.Psi)
    assert eq.pressure.equiv(eq_VMEC.pressure)
    assert eq.iota.equiv(eq_VMEC.iota)
    assert eq.current is None
    assert eq_VMEC.current is None
    assert eq.sym == eq_VMEC.sym

    # check against the DSHAPE bdry and profiles
    eq_example = desc.examples.get("DSHAPE")
    eq.change_resolution(L=eq_example.L, M=eq_example.M)
    np.testing.assert_allclose(eq.Rb_lmn, eq_example.Rb_lmn)
    np.testing.assert_allclose(eq.Zb_lmn, eq_example.Zb_lmn)
    np.testing.assert_allclose(eq.Psi, eq_example.Psi)
    np.testing.assert_allclose(eq.p_l, eq_example.p_l)
    # our example's iota is negative of this input files's
    np.testing.assert_allclose(eq.i_l, -eq_example.i_l)
    assert eq.sym == eq_example.sym


@pytest.mark.unit
def test_from_input_file_equilibrium_desc_vmec():
    """Test that from_input_file gives same eq for DESC and VMEC input files."""
    vmec_path = ".//tests//inputs//input.QSC_r2_5.5_vmec"
    desc_path = ".//tests//inputs//input.QSC_r2_5.5_desc"
    kwargs = {"L": 10, "M": 10, "N": 14}
    eq = Equilibrium.from_input_file(desc_path, **kwargs)
    with pytest.warns(UserWarning):
        eq_VMEC = Equilibrium.from_input_file(vmec_path, **kwargs)

    np.testing.assert_allclose(eq.R_lmn, eq_VMEC.R_lmn)
    np.testing.assert_allclose(eq.Z_lmn, eq_VMEC.Z_lmn)
    np.testing.assert_allclose(eq.L_lmn, eq_VMEC.L_lmn)
    np.testing.assert_allclose(eq.Rb_lmn, eq_VMEC.Rb_lmn)
    np.testing.assert_allclose(eq.Zb_lmn, eq_VMEC.Zb_lmn)
    np.testing.assert_allclose(eq.Ra_n, eq_VMEC.Ra_n)
    np.testing.assert_allclose(eq.Za_n, eq_VMEC.Za_n)
    np.testing.assert_allclose(eq.Psi, eq_VMEC.Psi)
    assert eq.pressure.equiv(eq_VMEC.pressure)
    assert eq.current.equiv(eq_VMEC.current)
    assert eq.iota is None
    assert eq_VMEC.iota is None
    assert eq.sym == eq_VMEC.sym

    if os.path.exists(".//tests//inputs//input.QSC_r2_5.5_vmec_desc"):
        os.remove(".//tests//inputs//input.QSC_r2_5.5_vmec_desc")


@pytest.mark.unit
def test_vmec_input_surface_threshold():
    """Test ."""
    path = ".//tests//inputs//input.QSC_r2_5.5_vmec"
    with pytest.warns(UserWarning, match="Detected multiple inputs"):
        surf_full = InputReader.parse_vmec_inputs(path)[-1]["surface"]
        surf_trim = InputReader.parse_vmec_inputs(path, threshold=1e-6)[-1]["surface"]
    assert surf_full.shape[0] > surf_trim.shape[0]
    assert surf_full.shape[1] == surf_trim.shape[1] == 5


class TestInputReader:
    """Tests for the InputReader class."""

    argv0 = []
    argv1 = ["nonexistent_input_file"]
    argv2 = ["./tests/inputs/MIN_INPUT"]

    @pytest.mark.unit
    def test_no_input_file(self):
        """Test an error is raised when no input file is given."""
        with pytest.raises(NameError):
            InputReader(cl_args=self.argv0)

    @pytest.mark.unit
    def test_nonexistant_input_file(self):
        """Test error is raised when nonexistent path is given."""
        with pytest.raises(FileNotFoundError):
            InputReader(cl_args=self.argv1)

    @pytest.mark.unit
    def test_min_input(self):
        """Test that minimal input is parsed correctly."""
        ir = InputReader(cl_args=self.argv2)
        assert ir.args.input_file[0] == self.argv2[0], "Input file name does not match"
        assert ir.input_path == str(
            pathlib.Path("./" + self.argv2[0]).resolve()
        ), "Path to input file is incorrect."
        # Test defaults
        assert ir.args.plot == 0, "plot is not default 0"
        assert ir.args.quiet is False, "quiet is not default False"
        assert ir.args.verbose == 1, "verbose is not default 1"
        assert ir.args.numpy is False, "numpy is not default False"
        assert (
            os.environ["DESC_BACKEND"] == "jax"
        ), "numpy environment variable incorrect with default argument"
        assert ir.args.version is False, "version is not default False"
        assert (
            len(ir.inputs[0]) == 26
        ), "number of inputs does not match number expected in MIN_INPUT"
        # test equality of arguments

    @pytest.mark.unit
    def test_np_environ(self):
        """Test setting numpy backend via environment variable."""
        argv = self.argv2 + ["--numpy"]
        InputReader(cl_args=argv)
        assert (
            os.environ["DESC_BACKEND"] == "numpy"
        ), "numpy environment variable incorrect on use"

    @pytest.mark.unit
    def test_quiet_verbose(self):
        """Test setting of quiet and verbose options."""
        ir = InputReader(self.argv2)
        assert (
            ir.inputs[0]["verbose"] == 1
        ), "value of inputs['verbose'] incorrect on no arguments"
        argv = self.argv2 + ["-v"]
        ir = InputReader(argv)
        assert (
            ir.inputs[0]["verbose"] == 2
        ), "value of inputs['verbose'] incorrect on verbose argument"
        argv = self.argv2 + ["-vv"]
        ir = InputReader(argv)
        assert (
            ir.inputs[0]["verbose"] == 3
        ), "value of inputs['verbose'] incorrect on double verbose argument"
        argv = self.argv2 + ["-q"]
        ir = InputReader(argv)
        assert (
            ir.inputs[0]["verbose"] == 0
        ), "value of inputs['verbose'] incorrect on quiet argument"

    @pytest.mark.unit
    def test_vacuum_objective_with_iota_yields_current(self):
        """Test that input file with vacuum objective always uses zero current."""
        input_path = ".//tests//inputs//HELIOTRON_vacuum"
        # load an input file with vacuum obj but also an iota profile specified
        with pytest.warns(UserWarning):
            ir = InputReader(input_path)
        # ensure that no profiles used
        assert "iota" not in ir.inputs[0].keys()
        assert "current" not in ir.inputs[0].keys()
        assert "pressure" not in ir.inputs[0].keys()
        assert ir.inputs[0]["objective"] == "force"

    @pytest.mark.unit
    def test_node_pattern_warning(self):
        """Test that a warning is thrown when trying to use a custom node pattern."""
        input_path = ".//tests//inputs//SOLOVEV_poincare"
        # load an input file with vacuum obj but also an iota profile specified
        with pytest.warns(UserWarning):
            ir = InputReader(input_path)
        assert "node_pattern" not in ir.inputs[0]


class MockObject:
    """Example object for saving/loading tests."""

    def __init__(self):
        self._io_attrs_ = ["a", "b", "c"]


@pytest.mark.unit
def test_writer_given_filename(writer_test_file):
    """Test writing to a given file by filename."""
    writer = hdf5Writer(writer_test_file, "w")
    assert writer.check_type(writer.target) is False
    assert writer.check_type(writer.base) is True
    assert writer._close_base_ is True
    writer.close()
    assert writer._close_base_ is False


@pytest.mark.unit
def test_writer_given_file(writer_test_file):
    """Test writing to given file instance."""
    f = h5py.File(writer_test_file, "w")
    writer = hdf5Writer(f, "w")
    assert writer.check_type(writer.target) is True
    assert writer.check_type(writer.base) is True
    assert writer._close_base_ is False
    assert writer._close_base_ is False
    f.close()


@pytest.mark.unit
def test_writer_close_on_delete(writer_test_file):
    """Test that files are closed when writer is deleted."""
    writer = hdf5Writer(writer_test_file, "w")
    with pytest.raises(OSError):
        newwriter = hdf5Writer(writer_test_file, "w")
    del writer
    newwriter = hdf5Writer(writer_test_file, "w")
    del newwriter


@pytest.mark.unit
def test_writer_write_dict(writer_test_file):
    """Test writing dictionary to hdf5 file."""
    thedict = {"1": 1, "2": 2, "3": 3}
    writer = hdf5Writer(writer_test_file, "w")
    writer.write_dict(thedict)
    with pytest.raises(SyntaxError):
        writer.write_dict(thedict, where="not a writable type")
    writer.close()
    f = h5py.File(writer_test_file, "r")
    for key in thedict.keys():
        assert key in f.keys()
        assert f[key][()] == thedict[key]
    f.close()
    reader = hdf5Reader(writer_test_file)

    dict1 = reader.read_dict()
    assert dict1 == thedict
    reader.close()


@pytest.mark.unit
def test_writer_write_list(writer_test_file):
    """Test writing list to hdf5 file."""
    thelist = ["1", 1, "2", 2, "3", 3]
    writer = hdf5Writer(writer_test_file, "w")
    writer.write_list(thelist)
    with pytest.raises(SyntaxError):
        writer.write_list(thelist, where="not a writable type")
    writer.close()
    reader = hdf5Reader(writer_test_file)

    list1 = reader.read_list()
    assert list1 == thelist
    reader.close()


@pytest.mark.unit
def test_writer_write_obj(writer_test_file):
    """Test writing objects to hdf5 file."""
    mo = MockObject()
    writer = hdf5Writer(writer_test_file, "w")
    # writer should throw runtime warning if any save_attrs are undefined
    with pytest.warns(RuntimeWarning):
        writer.write_obj(mo)
    writer.close()
    writer = hdf5Writer(writer_test_file, "w")
    for name in mo._io_attrs_:
        setattr(mo, name, name)
    writer.write_obj(mo)
    groupname = "initial"
    writer.write_obj(mo, where=writer.sub(groupname))
    writer.close()
    f = h5py.File(writer_test_file, "r")
    for key in mo._io_attrs_:
        assert key in f.keys()
    assert groupname in f.keys()
    initial = f[groupname]
    for key in mo._io_attrs_:
        assert key in initial.keys()
    f.close()


@pytest.mark.unit
def test_reader_given_filename(reader_test_file):
    """Test opening a reader with a given filename."""
    reader = hdf5Reader(reader_test_file)
    assert reader.check_type(reader.target) is False
    assert reader.check_type(reader.base) is True
    assert reader._close_base_ is True
    reader.close()
    assert reader._close_base_ is False


@pytest.mark.unit
def test_reader_given_file(reader_test_file):
    """Test opening a reader from a given file instance."""
    f = h5py.File(reader_test_file, "r")
    reader = hdf5Reader(f)
    assert reader.check_type(reader.target) is True
    assert reader.check_type(reader.base) is True
    assert reader._close_base_ is False
    assert reader._close_base_ is False
    f.close()


@pytest.mark.unit
def test_reader_read_obj(reader_test_file):
    """Test reading an object from hdf5 file."""
    mo = MockObject()
    reader = hdf5Reader(reader_test_file)
    reader.read_obj(mo)
    mo._io_attrs_ += "4"
    with pytest.warns(RuntimeWarning):
        reader.read_obj(mo)
    del mo._io_attrs_[-1]
    submo = MockObject()
    reader.read_obj(submo, where=reader.sub("subgroup"))
    for key in mo._io_attrs_:
        assert hasattr(mo, key)
        assert hasattr(submo, key)


@pytest.mark.unit
@pytest.mark.solve
def test_pickle_io(tmpdir_factory):
    """Test saving and loading equilibrium in pickle format."""
    tmpdir = tmpdir_factory.mktemp("desc_inputs")
    tmp_path = tmpdir.join("solovev_test.pkl")
    eqf = desc.examples.get("DSHAPE_CURRENT", "all")
    eqf.save(tmp_path, file_format="pickle")
    peqf = load(tmp_path, file_format="pickle")
    assert equals(eqf, peqf)


@pytest.mark.unit
@pytest.mark.solve
def test_ascii_io(tmpdir_factory):
    """Test saving and loading equilibrium in ASCII format."""
    tmpdir = tmpdir_factory.mktemp("desc_inputs")
    tmp_path = tmpdir.join("solovev_test.txt")
    eq1 = desc.examples.get("DSHAPE_CURRENT")
    eq1.iota = eq1.get_profile("iota", grid=LinearGrid(30, 16, 0)).to_powerseries(
        sym=True
    )
    write_ascii(tmp_path, eq1)
    with pytest.warns(UserWarning, match="not an even power series"):
        eq2 = read_ascii(tmp_path)
    assert np.allclose(eq1.R_lmn, eq2.R_lmn)
    assert np.allclose(eq1.Z_lmn, eq2.Z_lmn)
    assert np.allclose(eq1.L_lmn, eq2.L_lmn)


@pytest.mark.unit
def test_copy():
    """Test thing.copy() method of IOAble objects."""
    basis = FourierZernikeBasis(2, 2, 2)
    grid = LinearGrid(2, 2, 2)
    transform1 = Transform(grid, basis, method="direct1")
    transform2 = transform1.copy(deepcopy=False)

    assert transform1.basis is transform2.basis
    np.testing.assert_allclose(
        transform1.matrices["direct1"][0][0][0],
        transform2.matrices["direct1"][0][0][0],
        rtol=1e-10,
        atol=1e-10,
    )

    transform3 = transform1.copy(deepcopy=True)
    assert transform1.basis is not transform3.basis
    assert transform1.basis.equiv(transform3.basis)
    np.testing.assert_allclose(
        transform1.matrices["direct1"][0][0][0],
        transform3.matrices["direct1"][0][0][0],
        rtol=1e-10,
        atol=1e-10,
    )


@pytest.mark.unit
def test_save_none(tmpdir_factory):
    """Test that None attributes are saved/loaded correctly."""
    tmpdir = tmpdir_factory.mktemp("none_test")
    eq = Equilibrium()
    eq._iota = None
    eq.save(tmpdir + "none_test.h5")
    eq1 = load(tmpdir + "none_test.h5")
    assert eq1.iota is None


@pytest.mark.unit
def test_load_eq_without_current():
    """Test that loading an eq from DESC < 0.6.0 works correctly."""
    desc_no_current_path = ".//tests//inputs//DSHAPE_output_saved_without_current.h5"
    with pytest.warns(RuntimeWarning):
        eq = load(desc_no_current_path)[-1]
    assert eq.current is None


@pytest.mark.unit
def test_io_SplineMagneticField(tmpdir_factory):
    """Test saving/loading a SplineMagneticField works (tests dict saving)."""
    tmpdir = tmpdir_factory.mktemp("save_spline_field_test")
    tmp_path = tmpdir.join("spline_test.h5")

    R = np.linspace(1, 2, 2)
    Z = np.linspace(1, 2, 2)
    phi = np.linspace(1, 2, 2)

    field = SplineMagneticField.from_field(ToroidalMagneticField(R0=1, B0=1), R, phi, Z)

    field.save(tmp_path)
    field2 = load(tmp_path)

    for attr in field._io_attrs_:
        attr1 = getattr(field, attr)
        attr2 = getattr(field2, attr)

        if isinstance(attr1, str) or isinstance(attr1, bool):
            assert attr1 == attr2
        elif isinstance(attr1, dict):
            continue
        else:
            np.testing.assert_allclose(attr1, attr2, err_msg=attr)
    derivs1 = field._derivs
    derivs2 = field2._derivs
    for key in derivs1.keys():
        for key2 in derivs1[key].keys():
            np.testing.assert_allclose(derivs1[key][key2], derivs2[key][key2])


@pytest.mark.unit
def test_save_after_load(tmpdir_factory):
    """Test saving a DESC .h5 file after loading it (see gh #871)."""
    tmpdir = tmpdir_factory.mktemp("save_after_load_test")
    tmp_path = tmpdir.join("save_after_load_test.h5")
    eq = Equilibrium()
    eq.save(tmp_path)
    eq2 = load(tmp_path)
    # ensure can save again without an error being thrown due
    # to the .h5 file not being closed
    eq2.save(tmp_path)
    assert eq2.equiv(eq)


@pytest.mark.unit
def test_io_OmnigenousField(tmpdir_factory):
    """Test saving/loading an OmnigenousField works (tests dict saving)."""
    tmpdir = tmpdir_factory.mktemp("save_omnigenous_field_test")
    tmp_path = tmpdir.join("omnigenous_test.h5")

    field1 = OmnigenousField(
        L_B=1,
        M_B=4,
        L_x=0,
        M_x=1,
        N_x=1,
        NFP=4,
        helicity=(1, 4),
        B_lm=np.array([0.8, 0.9, 1.1, 1.2, 0, 0, 0, 0]),
        x_lmn=np.array([0, -np.pi / 8, 0, np.pi / 8, 0, np.pi / 4]),
    )

    field1.save(tmp_path)
    field2 = load(tmp_path)

    for attr in field1._io_attrs_:
        attr1 = getattr(field1, attr)
        attr2 = getattr(field2, attr)

        if isinstance(attr1, str) or isinstance(attr1, bool):
            assert attr1 == attr2
        elif hasattr(attr1, "_modes"):
            np.testing.assert_allclose(attr1.modes, attr2.modes, err_msg=attr)
        else:
            np.testing.assert_allclose(attr1, attr2, err_msg=attr)

    data1 = field1.compute(["|B|", "theta_B", "zeta_B"])
    data2 = field2.compute(["|B|", "theta_B", "zeta_B"])
    for key in data1.keys():
        np.testing.assert_allclose(data1[key], data2[key])
